"""
TextDistance.
Compute distance between sequences.
30+ algorithms, pure python implementation, common interface.
"""

# main package info
__title__ = 'TextDistance'
__version__ = '4.6.2'
__author__ = 'Gram (@orsinium)'
__license__ = 'MIT'


# version synonym
VERSION = __version__


# app
from .algorithms import *  # noQA
from .utils import *  # noQA
